#include <stdlib.h>
#include "CASL.h"
#include "syntax_tree.h"
#include "y.tab.h"

/* defined in code_exp.c */
extern void codeExp(Tree t);

/* defined in CASL.c */
char *newLabel();
void label(char *);
#define C0(c) C3(c, NULL, NULL, NULL)
#define C1(c, op1) C3(c, op1, NULL, NULL)
#define C2(c, op1, op2) C3(c, op1, op2 , NULL)
void C3(char *, char *, char *, char *);
char *d(int);
void comment(char *);
void dumpCode();
void putStringLiterals();

void initial_code() {
  label("P");
  C0(START); comment("initial code");
  C2(LAD, GR7, "DSTACK");
  C1(JUMP, "MAIN");
  C0(END);
  dumpCode();
}

void codeStmt(Tree);

void code(Tree t) {
  Tree s; Obj ob;
  switch (t->type) {
  case GVAR_DECL:
    for (s=t->left; s!=NUL; s=s->right) {
      switch (s->left->type) {
      case SCALAR:
        ob = s->left->left->obj;
        label(ob->name); C0(START); comment("global variable");
        C1(DS, d(1));
        C0(END);
        dumpCode();
        return;
      case ARRAY:
        ob = s->left->left->obj;
        label(ob->name); C0(START); comment("global array");
        C1(DS, d(ob->size));
        C0(END);
        dumpCode();
        return;
      }
    }
  case FUNC_DEF:
    ob = t->left->obj;
    label(ob->name); C0(START);
    C3(ST,GR7,d(ob->size+1),GR7);
    C3(LAD,GR7,d(ob->size+1),GR7); comment("end of function prefix");
    codeStmt(t->third);
    C3(LD,GR7,d(0),GR7); comment("function postfix");
    C0(RET);
    putStringLiterals();
    C0(END);
    dumpCode();
  }
}

void codeStmt(Tree t) {
  Tree s; Obj ob; char *l1, *l2;
  if (t==NUL) return;
  switch (t->type) {
  case LVAR_DECL:
    return;
  case EXP:
    codeExp(t->left);
    C1(POP,GR1); comment("postfix of expression as statement");
    return;
  case RETURN:
    if (t->left!=NUL) {
      codeExp(t->left);
      C1(POP,GR1);
    }
    C3(LD,GR7,d(0),GR7); comment("postfix of RETURN statement");
    C0(RET);
    return;
  case LIST:
    for (s=t; s!=NUL; s=s->right) {
      codeStmt(s->left);
    }
    return;
  case IF:
    l1 = newLabel();
    codeExp(t->left);
    C1(POP, GR1);
    C2(LD, GR1, GR1);
    C1(JZE, l1);
    codeStmt(t->right);
    if (t->third!=NUL) {
      l2 = newLabel();
      C1(JUMP,l2);
      label(l1); C1(DS, d(0)); comment("ELSE part");
      codeStmt(t->third);
      label(l2); C1(DS, d(0)); comment("end of ELSE part");
    } else {
      label(l1); C1(DS, d(0)); comment("end of IF statement");
    }
    return;
  case WHILE:
    l1 = newLabel(); l2 = newLabel();
    label(l1); C1(DS, d(0)); comment("WHILE statement");
    codeExp(t->left);
    C1(POP, GR1);
    C2(LD, GR1, GR1);
    C1(JZE, l2); comment("exit point of WHILE loop");
    codeStmt(t->right);
    C1(JUMP, l1); comment("end of WHILE loop");
    label(l2); C1(DS, d(0));
  }
}
